/***** sequencer.c ***************************************
 * Description: Program for simulating shotgun sequencing.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * Date: Tue Jun 22 17:30:49 2004.
 * License: GNU General Public
 *********************************************************/
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <math.h>
#include "sequence_data.h"
#include "eprintf.h"
#include "interface.h"
#include "ran.h"

void
run_sequencer (int fd, char *input_file, long *read_id, Args *args);

int
main (int argc, char *argv[])
{
  Args *args;			/* arguments */
  char *version;		/* program version */
  int i, fd;
  long read_id;
  int idum;
  FILE *fpra;

  version = "1.9";
  args = get_args (argc, argv);

  if (args->h == 1)
    {
      print_usage ();
      return 0;
    }
  else if (args->v)
    {
      print_version (version);
      return 0;
    }
  /* seed for random number generation */
  if (args->s != 0)
    idum = args->s;
  else if ((fpra = fopen ("random_seed.dat", "r")) != NULL)
    {
      fscanf (fpra, "%d", &idum);
      fclose (fpra);
    }
  else
    idum = -time (NULL);
  init_genrand (idum);

  read_id = 0;
  if (args->num_input_files == 0)
    {
      fd = 0;
      run_sequencer (fd, "stdin", &read_id, args);
    }
  else 
    {
      for(i = 0; i < args->num_input_files; i++)
	{
	  if ((fd = open (args->input_files[i], O_RDONLY, 0)) == 0)
	    eprintf ("ERROR [sequencer]: cannot open file %s for reading\n", args->input_files[i]);
	  run_sequencer (fd, args->input_files[i], &read_id, args);
	  close(fd);
	}
    }
  if (args->s == 0)
    {
      fpra = efopen ("random_seed.dat", "w");
      fprintf (fpra, "%d\n", (int) genrand_int32 ());
      fclose (fpra);
    }

  free (args);
  free (progname ());
  return 0;
}

void
run_sequencer (int fd, char *input_file, long *read_id, Args *args)
{
  Sequence *seq;		/* sequence data */
  float sd;			/* standard deviation */
  int residues;
  int sum, coverage, p, pf, len, printed_len, a, i, c, j;
  int seqcounter;
  int seqi, seq_ind, max_res;
  int revcounter = 0;
  int line_length;
  int *num_res, **profiles = NULL, *dic = NULL;
  FILE *fout;
  char *sequence[2], strand;
  char **forward, **reverse;
  int nuc, new_nuc;
  char *dic_str = "ACGTN";

  seq = NULL;
  line_length = args->L;
  seq = read_fasta (fd);
  fout = stdout;

  /* generate conveniently accessible forward & reverse strands */
  forward = (char **) emalloc (seq->num_seq * sizeof (char *));
  reverse = (char **) emalloc (seq->num_seq * sizeof (char *));
  num_res = (int *) emalloc (seq->num_seq * sizeof (int));
  forward[0] = seq->seq;
  num_res[0] = seq->borders[0];
  max_res = num_res[0];
  for (i = 1; i < seq->num_seq; i++)
    {
      forward[i] = seq->seq + seq->borders[i - 1] + 1;
      num_res[i] = seq->borders[i] - seq->borders[i - 1] - 1;
      if (num_res[i] > max_res)
	max_res = num_res[i];
      forward[i][num_res[i]] = '\0';
    }
  dic = get_restricted_dna_dictionary (dic);
  if (args->p)
    profiles = initialize_profiles (profiles, max_res);
  for (i = 0; i < seq->num_seq; i++)
    reverse[i] = revcomp_string (forward[i]);
  coverage = args->c * seq->num_nuc;
  sd = args->d;
  sum = 0;

  seqcounter = 0;
  sum = 0;
  seq_ind = 0;
  while (sum < coverage)
    {
      /* choose input sequence */
      seq_ind = genrand_real1 () * seq->num_seq;
      sequence[0] = forward[seq_ind];
      sequence[1] = reverse[seq_ind];
      residues = num_res[seq_ind];
      (*read_id)++;
      if(args->S) /* use fixed position */
	p = args->S - 1;
      else        /* generate random position */
	p = (int) (genrand_real1 () * residues);
      pf = p;
      /* get random sequence length */
      len = args->l + (rand_Normal () * sd) + 0.5;
      /* determine strand */
      if (!args->f)
	{
	  if (genrand_real1 () > 0.5){
	    seqi = 1;
	    a = residues - p;
	    len = (len > a) ? a : len;
	    p = residues - p - len;
	  }else
	    seqi = 0;
	}
      else
	seqi = 0;
      /* write sequence to output */
      i = 0;
      c = 0;
      if (args->r || len <= residues - p + 1)
	printed_len = len;
      else
	printed_len = residues - p;
      if (seqi){
	strand = 'r';
/* 	p++; */
      }else
	strand = 'f';
      for (j = 0; j < strlen(seq->headers[i]); j++)
	if(seq->headers[i][j] == ' ')
	  {
	    seq->headers[i][j] = '\0';
	    break;
	  }
      if (!args->p)
	{
	  fprintf (fout, ">Read#%d file=%s sequence=%s strand=%c start=%d len=%d\n",
		   (int)*read_id, input_file, seq->headers[seq_ind] + 1, strand, pf + 1, printed_len);
	}
      while (i < len && p + i < residues)
	{
	  if (c == line_length && !args->p)
	    {
	      fprintf (fout, "\n");
	      c = 0;
	    }
	  nuc = dic[(int) sequence[seqi][p + i]];
	  /* sequencing error */
	  if (genrand_real1 () <= args->E)
	    {
	      while ((new_nuc = (int) (genrand_real1 () * 4)) == nuc);
	      nuc = new_nuc;
	    }
	  if (!args->p)
	    fprintf (fout, "%c", dic_str[nuc]);
	  else
	    profiles[p + i][nuc]++;
	  i++;
	  c++;
	  sum++;
	}
      /* take care of circular genome */
      if (args->r)
	{
	  revcounter = 0;
	  while (i < len)
	    {
	      if (c == line_length && !args->p)
		{
		  fprintf (fout, "\n");
		  c = 0;
		}
	      nuc = dic[(int) sequence[seqi][revcounter]];
	      /* sequencing error */
	      if (genrand_real1 () <= args->E)
		{
		  while ((new_nuc = (int) (genrand_real1 () * 4)) == nuc);
		  nuc = new_nuc;
		}
	      if (!args->p)
		fprintf (fout, "%c", dic_str[nuc]);
	      else
		profiles[revcounter][nuc]++;
	      revcounter++;
	      i++;
	      c++;
	      sum++;
	    }
	}
      if (c != 0 && !args->p)
	fprintf (fout, "\n");
    }
  if (args->p)
    print_profiles (fout, profiles, max_res, args->m);
/*   fclose (fout); */
}
